import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.BufferedReader;
import java.io.FileReader;


/** HSQLDB Parser
 * Matt Mancuso
 * Andy Feeney
 * WPI MQP 
 * Last Modified: 21 April 2013
 * Based heavily on example code at 
 * http://hsqldb.org/doc/guide/apb.html 
 * by The HSQL Development Group
 */

/** This class is used to generate summary and header output files from a ZAP HSQLDB */
public class HeaderParser
{
	//The connection object
	Connection conn;
	
	//where the databases are located
	String directoryName;
	
	//What number database we're on
	int database;

	// Constructor, instantiates connection with databases in directory directoryName
	public HeaderParser(String directoryName) 
			throws SQLException
	{
		this.directoryName = directoryName;
		
		//Load jdbcDriver
		try 
		{
			Class.forName("org.hsqldb.jdbcDriver" );
		} 
		catch (Exception e) 
		{
			System.out.println("ERROR: failed to load HSQLDB JDBC driver.");
			e.printStackTrace();
			return;
		}

		//Connect to first database
		int database = 1;
		
		try
		{
			conn = DriverManager.getConnection("jdbc:hsqldb:file:"+directoryName+"/"+database+".session","sa","");
		}
		catch(Exception e)
		{
			System.out.println("ERROR: Cannot connect to "+directoryName+"/"+database+".session");
			e.printStackTrace();
			return;
		}
	}

	/* Main */
	public static void main(String[] args) 
			throws Exception 
	{
		//if no arguments specified
		if (args.length<1)
		{
			//print usage instructions
			System.out.println("USAGE: ServerTest.jar inputDB summaryOutput levelOfDetail [1|2|3] opt:headerOutput");
			return;
		}

		//initialize variable to null to detect errors later
		String inputURLs=null;
		String dbName=null;
		String summaryOut=null;
		String headerOut=null;
		Integer lod=(Integer)null;
		
		//iterate through arguments
		int arg=0;
		
		for (String s : args)
		{
			arg++;
			switch(arg)
			{
				//input URL file
				case 1:
					inputURLs=s;
					break;
				//database name 
				case 2:
					dbName=s;
					break;
				//summary output file
				case 3:
					summaryOut=s;
					break;
				//level of detail
				case 4:
					lod=Integer.parseInt(s);
					break;
				//header output file
				case 5:
					headerOut=s;
					break;
			}
		}

		//if generating a header file was specified but no output was given
		if (lod>1 && headerOut==null)
		{
			System.out.println("Must provide headerOut if specifying a LOD > 1");
			return;
		}

		//if a required field wasn't inputed properly
		if (inputURLs==null||dbName==null || summaryOut==null || lod==(Integer)null)
		{
			System.out.println("Insufficient arguments");
			return;
		}

		//create new connection
		HeaderParser s = new HeaderParser(dbName);
		
		//Regular Expressions
		Pattern textData = 
				Pattern.compile("Content-Type: (application/.*(javascript|json)|text)");
		Pattern contentType = Pattern.compile("Content-Type: (.*)");
		Pattern contentLength = Pattern.compile("Content-Length: (.*)");
		Pattern imageData = Pattern.compile("Content-Type: image/(png|gif|jpeg|jpg)");
		Pattern URLregex = Pattern.compile("^(https?://)?([a-z0-9.]+)\\/.*");

		try 
		{	
			//if lod > 1, i.e. user wants a header output
			//build header file
			if (lod>1)
			{	
				int headerIndex=0;								
				//Create writer for header output file
				BufferedWriter headerOutputFile = new BufferedWriter(new FileWriter(headerOut));

				while(true) 
				{
					//SQL Request
					List<Object> headers = s.query("SELECT URI,REQHEADER,RESHEADER,RESBODY FROM history WHERE STATUSCODE <> 0;");// WHERE NOT REQHEADER LIKE '%Content-length%';");
					//Iterator for list of objects returned by the SQL request
					Iterator<Object> it = headers.iterator();
					//While the list is not empty
					while (it.hasNext())
					{
						//get the URI (for image processing constraints)
						Object o = it.next();	
						//String URI = ((String)o);
						
						//Increment header count
						headerIndex++;
						
						//get Request Header
						o=it.next();
						
						//write item to file
						headerOutputFile.write((String)o);
						
						//get Response Header & write
						o=it.next();
						headerOutputFile.write((String)o);
						
						//if data was requested in the level-of-detail field
						if (lod==3)
						{
							//save response header to a temporary value
							//so we can advance to data in the list
							String responseHeader = ((String)o);
							
							//search header to see if it contains text data
							Matcher m = textData.matcher(responseHeader);
							
							//move to response body
							o=it.next();
							
							//if we found text data
							if (m.find())
							{
								//write text data to file directly
								String byteToString = new String((byte[])o);
								headerOutputFile.write(byteToString);
							}
							
							//if we found imageData
							m=imageData.matcher(responseHeader);
							
							//Suggestino: Add URI Limits to only operate on
							//content originating from ad sites, this will cut down a lot of unnecessary processing
/*							if (m.find())// && !URI.contains("adsense"))	//replace "adsense" with your list of ad domains
							{
								//System.out.println("Processing:\n"+responseHeader);
								//System.out.println(m.group(0));
								
								
								//save the image to a temporary file
								OutputStream tempImage = new BufferedOutputStream(new FileOutputStream("temp."+m.group(1)));
								tempImage.write((byte[])o);
								tempImage.close();
								
								//System.out.println("wrote image temp."+m.group(1));
								
								Runtime rt = Runtime.getRuntime();
								
								Process p = rt.exec("convert -compress none temp."+m.group(1)+" temp.tif");
								p.waitFor();
								p=rt.exec("tesseract temp.tif temp");
								p.waitFor();
								
								BufferedReader ocrdata = new BufferedReader(new FileReader("temp.txt"));
								
								String ocrtext=new String();
								String line;
								line=ocrdata.readLine();
								while (line != null){
									System.out.println(line);
									ocrtext+=line;
									line = ocrdata.readLine();
								}
								ocrdata.close();
								
								headerOutputFile.write(ocrtext+"\n");
							}
*/						}				
						else
						{
							//contains no data we're interested in
							//so skip over response body
							o=it.next();
						}
						
						//write delimiter to file
						headerOutputFile.write("\r\n\r\n----------------------------------------------------------  "+headerIndex+"\r\n\r\n");
					}
					if(s.lastDatabase()) 
					{
						break;
					}
					else 
					{
						s.nextDatabase();
					}
				}
				//close file
				headerOutputFile.close();
			}
			
			s.connectToDatabase(1);

			/* Build Summary File */
					
			//open output file
			BufferedWriter summaryOutputFile = new BufferedWriter(new FileWriter(summaryOut));
			
			//write file header
			//summaryOutputFile.write(String.format("%s %s %s %s %s %s %s\r\n\r\n","Request Number", "Req Time", "Res Time", "Code","Content Length","Content Type","URI"));

			//open input URL file for summary line parsing
			BufferedReader inputURLFile=new BufferedReader(new FileReader(inputURLs));

			//initialize variables
			String nextURL = null;
			String containmentURL = null;
			String rawNextURL = null;
			int totalSize=0;
			int totalTime=0;
			int totalRequests=0;
			int deadRequests=0;

			//read first line of URL list
			nextURL=inputURLFile.readLine();
			rawNextURL=nextURL;	
			containmentURL=nextURL;
			nextURL=inputURLFile.readLine();

			//strip protocol from string
			if(nextURL!=null)
			{
				Matcher m = URLregex.matcher(nextURL);
				if (m.find())
				{
					nextURL= m.group(2)!=null ? m.group(2) : nextURL;
				}
			}

			while(true) 
			{
				//SQL Request
				List<Object> summary = s.query("SELECT STATUSCODE,RESHEADER,METHOD,URI FROM history WHERE STATUSCODE <> 0;"); //WHERE NOT REQHEADER LIKE '%Content-length%';");
				//Iterator for SQL response object list
				Iterator<Object> it = summary.iterator();
				//while there are objects in the list to process
				while (it.hasNext())
				{
					//Get responsecode
					Object o = it.next();
					int rescode = ((int)o);
					
					//increment number of requests
					totalRequests++;
					
					//get response header
					o=it.next();
					//System.out.println((String)o);
					
					//check for Content-Length field
					Matcher m = contentLength.matcher((String)o);
					String conlen=null;
					
					//if there is a Content-Length
					if (m.find())
					{
						if (m.group(1)!=null)
						{
							//pull out Content-Length value from regex match
							conlen = m.group(1);
						}
						
						//ZAP Errors Record as dummy responses w/ response code 504 & length 63
						//if we encounter these we want to count it as a dead request
						//and not count the size of the item in the total size
						if(rescode==504 && Integer.parseInt(conlen)==63)
						{
							deadRequests++;
						}
						else
						{
							//no error, so just add in the size to the total
							totalSize+=Integer.parseInt(conlen);
						}
					}
					else //no Content-Length, so just add in a placeholder value
					{
						conlen="0";
					}
					
					//search header for Content-Type field
					m=contentType.matcher((String)o);
					String contype=null;
					
					//if there is a Content-Type
					if (m.find())
					{
						if (m.group(1)!=null)
						{
							//pull out Content-Type value from regex match
							contype=m.group(1);
							
							if (contype.indexOf(";")>0)
							{
								contype=contype.substring(0,contype.indexOf(";"));
							}
							
						}
					}
					else //no Content-Type, so add a placeholder value
					{
						contype="N/A";
					}
					
					//get Method
					o=it.next();
					
					//Uncomment this next line to have a method variable with the value of the HTTP method (GET|POST|etc)
					//String method=((String)o);	
					
					//get URI
					o=it.next();
					String URI=((String)o);
						
					//if this is a new top level domain (in the input list)
					//then print a summary and reset
					
					if (nextURL != null && URI.indexOf(nextURL)>=0)
					{
						if (!containmentURL.startsWith("http://"))
						{
							containmentURL="http://"+containmentURL;
						}
						if (!containmentURL.endsWith("/"))
						{
							containmentURL=containmentURL+"/";
						}
						
						//write summary line
						String summaryLine = String.format("summary: %s %d %d %d %d\r\n\r\n",containmentURL,totalSize,totalTime,totalRequests,deadRequests);
						System.out.println(summaryLine);
						
						summaryOutputFile.write(summaryLine);
	
						//reset variables
						totalSize=0;
						totalTime=0;
						totalRequests=1;
						deadRequests=0;
						
						//advance in the URL file
						containmentURL=rawNextURL;
						
						nextURL=inputURLFile.readLine();
						if(nextURL==null)
						{
							nextURL="END_OF_URL_LIST";
						}
						//strip protocol from string
						else
						{
							nextURL=nextURL.substring(nextURL.indexOf("://")+3);
						}
					}
					//write final item
					String line = String.format("%d 0 0 %d %s %s %s\r\n",totalRequests,rescode,conlen,contype,URI);
					summaryOutputFile.write(line);
				}
				//System.out.println(s.database);
				if(s.lastDatabase()) 
				{
					break;
				}
				else 
				{
					s.nextDatabase();
				}
			}
			
			if (!containmentURL.startsWith("http://"))
			{
				containmentURL="http://"+containmentURL;
			}
			if (!containmentURL.endsWith("/"))
			{
				containmentURL=containmentURL+"/";
			}
			
			//write final summary line
			String summaryLine = String.format("summary: %s %d %d %d %d\r\n\r\n",containmentURL,totalSize,totalTime,totalRequests,deadRequests);
			summaryOutputFile.write(summaryLine);

			//close files
			summaryOutputFile.close();
			inputURLFile.close();
		}
		catch (SQLException e) 
		{
			e.printStackTrace();
			return;
		}

		//kill connection
		s.shutdown();
	}
	
	//returns whether currently indexed database is the last
	public boolean lastDatabase() 
	{
		return !(new File(directoryName+"/"+(database+1)+".session").exists());
	}
	
	//connect to next database
	public void nextDatabase() throws SQLException 
	{
		shutdown();
		connectToDatabase(database+1);
	}
	
	//connect to specified database
	public void connectToDatabase(int database) 
	{
		this.database = database;
		try
		{
			conn = DriverManager.getConnection("jdbc:hsqldb:file:"+directoryName+"/"+database+".session","sa","");
		}
		catch(Exception e)
		{
			System.out.println("ERROR: Cannot connect to "+directoryName+"/"+database+".session");
			e.printStackTrace();
			return;
		}
	}

	/* Safely shutdown database and close connection */
	public void shutdown() 
			throws SQLException 
	{

		Statement st = conn.createStatement();

		//Give DB Command to exit and save operations
		st.execute("SHUTDOWN");

		//close connection
		conn.close();    	
	}

	//use for SQL command SELECT
	public synchronized List<Object> query(String expression) 
			throws SQLException 
	{
		Statement st = null;
		ResultSet rs = null;

		st = conn.createStatement();         
		// statement objects can be reused with
		// repeated calls to execute but we
		// choose to make a new one each time
		
		rs = st.executeQuery(expression);    // run the query

		// do something with the result set.
		ResultSet rscopy = rs;
		st.close();    
		return dump(rscopy);
	}

	//wrapper for requests, HSQL acts a little different from standard SQL
	public List<Object> dump(ResultSet rs) 
			throws SQLException 
	{
		// the order of the rows in a cursor
		// are implementation dependent unless you use the SQL ORDER statement
		ResultSetMetaData meta   = rs.getMetaData();
		int               colmax = meta.getColumnCount();
		int               i;
		Object            o = null;

		List<Object> oList = new LinkedList<Object>();

		// the result set is a cursor into the data.  You can only
		// point to one row at a time
		// assume we are pointing to BEFORE the first row
		// rs.next() points to next row and returns true
		// or false if there is no next row, which breaks the loop
		for (; rs.next(); ) 
		{
			for (i = 0; i < colmax; ++i) 
			{
				o = rs.getObject(i + 1);    // In SQL the first column is indexed
				oList.add(o);
			}
		}
		return oList;
	}
}
